/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this String except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.crypto;

import org.apache.commons.lang3.StringUtils;

import java.io.File;
import java.math.BigInteger;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Optional;

public abstract class SgdCryptoUtilities {
    public final static String OID_SGD1_HSM = "1.2.276.0.76.4.219";
    public final static String OID_SGD2_HSM = "1.2.276.0.76.4.220";
    public final static String OID_SGD = "1.2.276.0.76.4.221";
    public final static String OID_SGD_HSM_AUT = "1.2.276.0.76.4.214";
    public final static String OID_POLICY_GEM_OR_CP = "1.2.276.0.76.4.163";
    public final static String OID_EGK_AUT_ALT = "1.2.276.0.76.4.212";

    private static SgdCryptoUtilities instance;

    public static SgdCryptoUtilities getInstance() {
        if (instance == null) {
            throw new NullPointerException("No SgdCryptoUtilities provided! Please call setInstance() before calling SGD-Methods");
        }
        return instance;
    }

    public static void setInstance(SgdCryptoUtilities newInstance) {
        instance = newInstance;
    }

    public static String extractKVNRfromSubjectDN(String subjectDN) {
        String[] dn = StringUtils.split(subjectDN, ",");
        for (int i = 0; i < dn.length; i++) {
            dn[i] = dn[i].trim();
            String v = dn[i].replace(" ", "");
            if (v.startsWith("OU")) {
                if (v.length() == 13) {
                    return v.substring(3);
                }
            }
        }
        return null;
    }

    public abstract Optional<String> extractTelematikIdFromCertificate(X509Certificate certificate);

    public abstract X509Certificate loadX509CertificateFromP12File(String p12File, String p12Password);

    public abstract byte[] convertB64StringToBytes(String b64String);

    public abstract String convertBytesToB64String(byte[] encoded);

    public abstract X509Certificate loadX509CertificateFromCrtFile(String crtFile);

    public abstract PrivateKey loadPrivateKeyFromP12File(String p12File, String p12Password, String entryPassword);

    public abstract byte[] createSignatureUsingEcdsa(byte[] toBeSignedData, PrivateKey signerKey);

    public abstract KeyPair loadKeypairFromP12File(String p12File, String p12Password, String entryPassword);

    public abstract KeyPair generateEccKeyPair();

    public abstract KeyPair generateEccKeyPair(String curveName);

    public abstract byte[] parseHexStringToByteArray(String hexString);

    public abstract String convertByteArrayToHexString(byte[] byteArray);

    public abstract boolean doesCertificateContainAdmissionsExtensionOid(X509Certificate certificate, String admissionsOid);

    public abstract boolean isSgdHsmCertificateInTsl(X509Certificate certificate);

    public abstract String extractPublicPointCoordinatesFromPublicKey(PublicKey publicKey);

    public abstract String convertToLowerCaseHexWith64ByteLength(BigInteger value);

    public abstract String convertToLowerCaseHexWith64ByteLength(byte[] rawBytes);

    public abstract byte[] performHkdf(byte[] ikm, byte[] info, int length);

    public abstract void verifySignatureAndThrowExceptionWhenFail(byte[] payload, PublicKey publicKey, byte[] signatureInQuestion) throws SignatureException;

    public abstract void verifyEcdsaSignatureAndThrowExceptionWhenFail(byte[] toBeSignedBytes, PublicKey publicKey, byte[] signatureBytes) throws SignatureException;

    public abstract void verifyRsaSsaPssSignatureAndThrowExceptionWhenFail(byte[] toBeSignedBytes, PublicKey publicKey, byte[] signatureBytes) throws SignatureException;

    public abstract X509Certificate parseX509CertificateFromBytes(byte[] sgdCertificateRawBytes);

    public abstract byte[] deriveKeyAndEncrypt(PrivateKey senderPrivate, PublicKey receiverPublic, byte[] payload);

    public abstract byte[] deriveAesEncryptionKey(PrivateKey selfPrivateKey, PublicKey otherPublicKey);

    public abstract byte[] encryptUsingAesGcm(byte[] input, byte[] key, byte[] associatedData);

    public abstract byte[] decryptUsingAesGcm(byte[] encMessage, byte[] key, byte[] associatedData);

    public abstract PublicKey extractPublicKeyFromPublicPointCoordinates(String xString, String yString, String curveName);

    public abstract X509Certificate parseX509CertificateFromB64String(String certificate);

    public abstract void verifyOcspSignatureAndThrowExceptionWhenFail(byte[] ocspResponse, List<X509Certificate> possibleOcspSignerCertificates);

    public abstract byte[] signRsa(byte[] message, PrivateKey prk);

    public abstract byte[] signEcdsa(byte[] message, PrivateKey prk);

    public abstract byte[] deriveKeyAndDecrypt(PrivateKey privateKey, PublicKey publicKey, byte[] cipher);
}
