/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.domain.sgdCommunication.vectors;

import de.gematik.ti.sgd.components.SchluesselGenerierungsDienst;
import de.gematik.ti.sgd.domain.DerivationMode;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import org.apache.commons.lang3.StringUtils;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.SgdServerUtils.perform18003Conversion;

public class SgdVector {
    /////////////////////////////////// R1 Initial ////////////////////////////////////////////
    private static final AssociatedDataFunction R1_INITIAL_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (StringUtils.isEmpty(kvnr)) {
            throw new SgdServerException("FAIL (Empty KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (!StringUtils.equals(sgdVector.getParts().get(1), kvnr)) {
            throw new SgdServerException("FAIL (Missmatching KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return "r1:" + rnd + ":" + sgdVector.getParts().get(1) + ":" + keyRepository.getCurrentDerivationKeyName();
    };
    private static final DerivationKeyFunction R1_INITIAL_KEY_FUNCTION = (sgdVector, keyRepository) -> keyRepository.getCurrentDerivationKeyValue();
    /////////////////////////////////// R1 Repeated ////////////////////////////////////////////
    private static final AssociatedDataFunction R1_REPEATED_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (StringUtils.isEmpty(kvnr)) {
            throw new SgdServerException("FAIL (Empty KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (!StringUtils.equals(sgdVector.getParts().get(2), kvnr)) {
            throw new SgdServerException("FAIL (Missmatching KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (sgdVector.getParts().get(1).length() != 64) {
            throw new SgdServerException("FAIL (RND length missmatch!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return sgdVector.toString();
    };
    private static final DerivationKeyFunction R1_REPEATED_KEY_FUNCTION = (sgdVector, keyRepository) ->
            keyRepository.getDerivationKeyByName(sgdVector.getKeyDerivationName())
                    .orElseThrow(() -> new SgdServerException("FAIL (Unknown Derivator-Key specified)", SgdResponseStatus.KEY_DERIVATION_FAIL));
    /////////////////////////////////// R2 Initial ////////////////////////////////////////////
    private static final AssociatedDataFunction R2_INITIAL_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (StringUtils.isEmpty(kvnr)) {
            throw new SgdServerException("FAIL (Empty KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (StringUtils.isEmpty(sgdVector.getParts().get(1))) {
            throw new SgdServerException("FAIL (Empty s[1] discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return "r2:" + rnd + ":" + kvnr + ":" + sgdVector.getParts().get(1) + ":" + keyRepository.getCurrentDerivationKeyName();
    };
    private static final DerivationKeyFunction R2_INITIAL_KEY_FUNCTION = (sgdVector, keyRepository) -> keyRepository.getCurrentDerivationKeyValue();
    /////////////////////////////////// R2 Repeated ////////////////////////////////////////////
    private static final AssociatedDataFunction R2_REPEATED_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (sgdVector.getParts().get(1).length() != 64) {
            throw new SgdServerException("FAIL (RND length missmatch!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (StringUtils.isEmpty(sgdVector.getParts().get(2))) {
            throw new SgdServerException("FAIL (empty s[2] discovered, Check 12.3 failed!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        if (StringUtils.isEmpty(kvnr) && StringUtils.isEmpty(telematikId)) {
            throw new SgdServerException("FAIL (KVNR and TELEMATIK_ID empty, Check 12.5 failed!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        String convertedTelematikId = perform18003Conversion(telematikId);
        if ((!sgdVector.getParts().get(3).equals(convertedTelematikId))
                && (!sgdVector.getParts().get(3).equals(kvnr))) {
            throw new SgdServerException("FAIL (Check 12.7 failed!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return sgdVector.toString();
    };
    private static final DerivationKeyFunction R2_REPEATED_KEY_FUNCTION = (sgdVector, keyRepository) ->
            keyRepository.getDerivationKeyByName(sgdVector.getKeyDerivationName())
                    .orElseThrow(() -> new SgdServerException("FAIL (Unknown Derivator-Key '" + sgdVector.getKeyDerivationName() + "' specified)", SgdResponseStatus.KEY_DERIVATION_FAIL));
    /////////////////////////////////// R3 Initial ////////////////////////////////////////////
    private static final AssociatedDataFunction R3_INITIAL_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (StringUtils.isEmpty(kvnr)) {
            throw new SgdServerException("FAIL (Empty KVNR discovered!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return "r3:" + rnd + ":" + sgdVector.getParts().get(2) + ":" + kvnr + ":" + sgdVector.getParts().get(1) + ":" + keyRepository.getCurrentDerivationKeyName();
    };
    private static final DerivationKeyFunction R3_INITIAL_KEY_FUNCTION = (sgdVector, keyRepository) -> keyRepository.getCurrentDerivationKeyValue();
    /////////////////////////////////// R3 Repeated ////////////////////////////////////////////
    private static final AssociatedDataFunction R3_REPEATED_AD_FUNCTION = (sgdVector, rnd, keyRepository, kvnr, telematikId) -> {
        if (sgdVector.getParts().get(1).length() != 64) {
            throw new SgdServerException("FAIL (RND length missmatch!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        String convertedTelematikId = perform18003Conversion(telematikId);
        if (!sgdVector.getParts().get(4).equals(convertedTelematikId)) {
            throw new SgdServerException("FAIL (Check 14.5 failed!)", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return sgdVector.toString();
    };
    private static final DerivationKeyFunction R3_REPEATED_KEY_FUNCTION = (sgdVector, keyRepository) ->
            keyRepository.getDerivationKeyByName(sgdVector.getKeyDerivationName())
                    .orElseThrow(() -> new SgdServerException("FAIL (Unknown Derivator-Key '" + sgdVector.getKeyDerivationName() + "' specified)", SgdResponseStatus.KEY_DERIVATION_FAIL));

    private final List<String> parts;
    private final DerivationMode derivationMode;
    private AssociatedDataFunction associatedDataFunction;
    private DerivationKeyFunction derivationKeyFunction;
    private boolean initialVector = false;

    public SgdVector(String content) {
        this(Arrays.asList(content.split(":")));
    }

    public SgdVector(List<String> content) {
        this.parts = content;
        this.derivationMode = DerivationMode.fromMessagePart(parts.get(0));

        if (derivationMode == DerivationMode.R1) {
            if (this.getParts().size() == 2) {
                initialVector = true;
                associatedDataFunction = R1_INITIAL_AD_FUNCTION;
                derivationKeyFunction = R1_INITIAL_KEY_FUNCTION;
            } else if (this.getParts().size() == 4) {
                initialVector = false;
                associatedDataFunction = R1_REPEATED_AD_FUNCTION;
                derivationKeyFunction = R1_REPEATED_KEY_FUNCTION;
            } else {
                throw new SgdServerException("Encountered ill-formatted content in R1-Sgd-Vector: Expected '" + content + "' to contain either 2 or 4 parts", SgdResponseStatus.KEY_DERIVATION_FAIL);
            }
        } else if (derivationMode == DerivationMode.R2) {
            if (this.getParts().size() == 2) {
                associatedDataFunction = R2_INITIAL_AD_FUNCTION;
                derivationKeyFunction = R2_INITIAL_KEY_FUNCTION;
                initialVector = true;
            } else if (this.getParts().size() == 5) {
                associatedDataFunction = R2_REPEATED_AD_FUNCTION;
                derivationKeyFunction = R2_REPEATED_KEY_FUNCTION;
                initialVector = false;
            } else {
                throw new SgdServerException("Encountered ill-formatted content in R2-Sgd-Vector: Expected '" + content + "' to contain either 2 or 5 parts", SgdResponseStatus.KEY_DERIVATION_FAIL);
            }
        } else if (derivationMode == DerivationMode.R3) {
            if (this.getParts().size() == 3) {
                associatedDataFunction = R3_INITIAL_AD_FUNCTION;
                derivationKeyFunction = R3_INITIAL_KEY_FUNCTION;
                initialVector = true;
            } else if (this.getParts().size() == 6) {
                associatedDataFunction = R3_REPEATED_AD_FUNCTION;
                derivationKeyFunction = R3_REPEATED_KEY_FUNCTION;
                initialVector = false;
            } else {
                throw new SgdServerException("Encountered ill-formatted content in R3-Sgd-Vector: Expected '" + content + "' to contain either 3 or 6 parts", SgdResponseStatus.KEY_DERIVATION_FAIL);
            }
        }
    }

    public static SgdVector createForR1InitialDerivation(String kvnr) {
        return new SgdVector("r1:" + kvnr);
    }

    public static SgdVector createForR1RepeatedDerivation(String rnd, String kvnr, String keyDerivationName) {
        return new SgdVector("r1:" + rnd + ":" + kvnr + ":" + keyDerivationName);
    }

    public static SgdVector createForR1RepeatedDerivation(String kvnr, String keyDerivationName) {
        return new SgdVector("r1:" + SchluesselGenerierungsDienst.generate256BitRandomString() + ":" + kvnr + ":" + keyDerivationName);
    }

    public static SgdVector createForR2InitialDerivation(String telematikId) {
        return new SgdVector("r2:" + telematikId);
    }

    public static SgdVector createForR2RepeatedDerivation(String rnd, String accountKvnr, String representativeKvnr, String keyDerivationName) {
        return new SgdVector("r2:" + rnd + ":" + accountKvnr + ":" + representativeKvnr + ":" + keyDerivationName);
    }

    public static SgdVector createForR2RepeatedDerivation(String accountKvnr, String representativeKvnr, String keyDerivationName) {
        return new SgdVector("r2:" + SchluesselGenerierungsDienst.generate256BitRandomString() + ":" + accountKvnr + ":" + representativeKvnr + ":" + keyDerivationName);
    }

    public static SgdVector createForR3InitialDerivation(String telematikId, String accountKvnr) {
        return new SgdVector("r3:" + telematikId + ":" + accountKvnr);
    }

    public static SgdVector createForR3RepeatedDerivation(String rnd, String accountKvnr, String representativeKvnr, String telematikId, String keyDerivationName) {
        return new SgdVector("r3:" + rnd + ":" + accountKvnr + ":" + representativeKvnr + ":" + telematikId + ":" + keyDerivationName);
    }

    public static SgdVector createForR3RepeatedDerivation(String accountKvnr, String representativeKvnr, String telematikId, String keyDerivationName) {
        return new SgdVector("r3:" + SchluesselGenerierungsDienst.generate256BitRandomString() + ":" + accountKvnr + ":" + representativeKvnr + ":" + telematikId + ":" + keyDerivationName);
    }

    public List<String> getParts() {
        return parts;
    }

    @Override
    public String toString() {
        return parts.stream()
                .collect(Collectors.joining(":"));
    }

    public String getType() {
        return getParts().get(0);
    }

    public String getRnd() {
        if (initialVector) {
            throw new SgdServerException("Could not give RND for initial " + derivationMode + " vector (content is '" + toString() + "')", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return getParts().get(1);
    }

    public Optional<String> getOwnerKvnr() {
        if (derivationMode == DerivationMode.R1) {
            return Optional.of(getParts().get(initialVector ? 1 : 2));
        } else if (derivationMode == DerivationMode.R2) {
            if (initialVector){
                return Optional.empty();
            } else {
                return Optional.of(getParts().get(2));
            }
        } else {
            return Optional.of(getParts().get(2));
        }
    }

    public Optional<String> getAuthorizedId() {
        if (derivationMode == DerivationMode.R1) {
            return getOwnerKvnr();
        } else if (derivationMode == DerivationMode.R2) {
            if (initialVector){
                return Optional.of(getParts().get(1));
            } else {
                return Optional.of(getParts().get(3));
            }
        } else {
            if (initialVector){
                return Optional.of(getParts().get(1));
            } else {
                return Optional.of(getParts().get(4));
            }
        }
    }

    public String getKeyDerivationName() {
        if (initialVector) {
            throw new SgdServerException("Could not give keyDerivationName for initial " + derivationMode + " vector (content is '" + toString() + "')", SgdResponseStatus.KEY_DERIVATION_FAIL);
        }
        return getParts().get(parts.size() - 1);
    }

    public boolean isInitialVector() {
        return initialVector;
    }

    public DerivationMode getDerivationMode() {
        return derivationMode;
    }

    public AssociatedDataFunction getAssociatedDataFunction() {
        return associatedDataFunction;
    }

    public DerivationKeyFunction getDerivationKeyFunction() {
        return derivationKeyFunction;
    }

    public interface AssociatedDataFunction {
        String getAssociatedData(SgdVector sgdVector, String rnd, DerivationKeyRepository keyRepository, String kvnr, String telematikId);
    }

    public interface DerivationKeyFunction {
        byte[] getDerivationKey(SgdVector sgdVector, DerivationKeyRepository keyRepository);
    }

    public interface DerivationKeyRepository {
        String getCurrentDerivationKeyName();

        byte[] getCurrentDerivationKeyValue();

        Optional<byte[]> getDerivationKeyByName(String keyName);
    }
}
