/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.domain.util;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.RecordAndContextKey;
import de.gematik.ws.fa.phr.v1.PHRKeyType;
import de.gematik.ws.fd.phrs.authorizationservice.v1.EncryptedKeyContainerType;

import javax.xml.bind.*;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class PHRUtils {
    public static final String AES_256_GCM_ALGORITHM_IDENTIFIER = "http://www.w3.org/2009/xmlenc11#aes256-gcm";
    public static final QName PHR_KEY_ELEMENT_NAME = new QName("http://ws.gematik.de/fa/phr/v1.1", "PHRKey");
    public static final QName ENCRYPTED_KEY_CONTAINER_ELEMENT_NAME = new QName("http://ws.gematik.de/fd/phrs/AuthorizationService/v1.1", "EncryptedKeyContainer");

    public static PHRKeyType decipherContainerToPHRKeyType(EncryptedKeyContainerType input, byte[] sgd1Key) {
        final byte[] phrKeyRaw;
        try {
            phrKeyRaw = SgdCryptoUtilities.getInstance().decryptUsingAesGcm(input.getCiphertext(), sgd1Key, DatatypeConverter.parseBase64Binary(input.getAssociatedData()));
        } catch (Exception e) {
            throw new RuntimeException("Error while deciphering EncryptedKeyContainer", e);
        }
        return unmarshallPhrKeyType(phrKeyRaw);
    }

    public static EncryptedKeyContainerType performFirstDecriptionStep(EncryptedKeyContainerType input, byte[] sgd2Key, byte[] associatedData) {
        final byte[] containerRaw = SgdCryptoUtilities.getInstance().decryptUsingAesGcm(input.getCiphertext(), sgd2Key, associatedData);
        return unmarshallEncryptedKeyContainer(containerRaw);
    }

    public static PHRKeyType decipherTwiceFromContainer(EncryptedKeyContainerType encryptedKeyContainer, byte[] sgd1KeyBytes, byte[] sgd2KeyBytes, byte[] outerAssociatedData) {
        EncryptedKeyContainerType intermediareContainer = performFirstDecriptionStep(encryptedKeyContainer, sgd2KeyBytes, outerAssociatedData);
        return PHRUtils.decipherContainerToPHRKeyType(intermediareContainer, sgd1KeyBytes);
    }

    public static byte[] marshall(Object value, QName rootElementName) {
        try {
            return marshal(value, rootElementName);
        } catch (JAXBException | IOException e) {
            throw new RuntimeException("Error while marshalling", e);
        }
    }

    public static RecordAndContextKey readRecordAndContextKeyFromPhrKeyXml(byte[] xmlBytes) {
        PHRKeyType phrKeyType;
        try {
            phrKeyType = unmarshallPhrKeyType(xmlBytes);
        } catch (Exception e) {
            throw new RuntimeException("Error while reading PHRKey", e);
        }

        return RecordAndContextKey.fromPHRKeyType(phrKeyType);
    }

    private static PHRKeyType unmarshallPhrKeyType(byte[] inBytes) {
        try {
            JAXBContext context = JAXBContext.newInstance(de.gematik.ws.fa.phr.v1.ObjectFactory.class);
            Unmarshaller unmarshaller = context.createUnmarshaller();
            ByteArrayInputStream is = new ByteArrayInputStream(inBytes);

            final JAXBElement<PHRKeyType> unmarshal = (JAXBElement<PHRKeyType>) unmarshaller.unmarshal(is);
            return unmarshal.getValue();
        } catch (JAXBException e) {
            throw new RuntimeException("Error during unmarshalling", e);
        }
    }
    
    public static EncryptedKeyContainerType unmarshallEncryptedKeyContainer(byte[] inBytes) {
        try {
            return unmarshal(inBytes, EncryptedKeyContainerType.class);
        } catch (JAXBException e) {
            throw new RuntimeException("Error during unmarshalling", e);
        }
    }

    public static byte[] buildOuterAesGcmAdFromOuterAssociatedData(String associatedData) {
        return Stream.of(associatedData.split("\\s+"))
                .map(String::trim)
                .map(b64EncodedStr -> DatatypeConverter.parseBase64Binary(b64EncodedStr))
                .map(String::new)
                .collect(Collectors.joining())
                .getBytes();
    }

    public static <T extends Object> T unmarshal(byte[] data, Class<T> targetClass) throws JAXBException {
        JAXBContext jaxbContext = JAXBContext.newInstance(targetClass);
        Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
        ByteArrayInputStream is = new ByteArrayInputStream(data);
        Source source = new StreamSource(is);
        JAXBElement jaxbElement = unmarshaller.unmarshal(source, targetClass);
        return (T) jaxbElement.getValue();
    }

    public static byte[] marshal(Object object, QName rootElementName) throws JAXBException, IOException {
        JAXBElement element = new JAXBElement(rootElementName, object.getClass(), object);
        return marshal(element, (object.getClass()));
    }

    public static byte[] marshal(JAXBElement element, Class myclass) throws JAXBException, IOException {
        JAXBContext context = JAXBContext.newInstance(myclass);
        Marshaller marshaller = context.createMarshaller();
        marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        marshaller.marshal(element, baos);
        baos.close();
        return baos.toByteArray();
    }
}
