/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.SgdInstance;
import de.gematik.ti.sgd.domain.json.GetPublicKeyResponse;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import de.gematik.ti.sgd.exceptions.client.InvalidCertificateException;
import de.gematik.ti.sgd.exceptions.client.InvalidSignatureException;
import org.junit.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

// A_18024, gemSpec_SGD_ePA
public class GetPublicKeyVerificationTest {

    private static final String P_12_PASSWORD = "00";
    private static final String PUBLIC_KEY_SPEC = "This is my Public Key Spec";
    private static SgdCryptoUtilities cryptoUtilities;
    private static List<PublicKey> SIGNER_LIST;

    static {
        cryptoUtilities = new SgdTestCryptoUtilities();

        SIGNER_LIST = extractPublicKeysFromListOfCertificateFilenames(Arrays.asList(
                "src/test/resources/keys/ca/GEM.KOMP-CA17-TEST-ONLY.crt",
                "src/test/resources/keys/ca/GEM.KOMP-CA10-TEST-ONLY.crt"));
    }

    private final String SGD_CERT_FILENAME = "src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.crt";
    private final String SGD_P12_FILENAME = "src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12";

    @Test
    public void responseContainsUnparseableCertificate_ShouldThrowInvalidCertificateException() {
        GetPublicKeyResponse sgdResponse = GetPublicKeyResponse.builder()
                .certificate("1234")
                .build();

        assertThatThrownBy(() -> SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(sgdResponse, null, SgdInstance.SGD_1))
                .isInstanceOf(InvalidCertificateException.class);
    }

    @Test
    public void validCertificateButInvalidSignature_ShouldThrowInvalidSignatureException() throws Exception {
        final X509Certificate certificate = cryptoUtilities.loadX509CertificateFromCrtFile(SGD_CERT_FILENAME);

        GetPublicKeyResponse sgdResponse = GetPublicKeyResponse.builder()
                .certificate(cryptoUtilities.convertBytesToB64String(certificate.getEncoded()))
                .signature("1234")
                .publicKeyEcies(new PublicKeyEcies("blub blab bli"))
                .build();

        assertThatThrownBy(() -> SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(sgdResponse, SIGNER_LIST, SgdInstance.SGD_1))
                .isInstanceOf(InvalidSignatureException.class);
    }

    @Test
    public void happyPath_noExceptionExpected() throws Exception {
        final X509Certificate certificate = cryptoUtilities.loadX509CertificateFromP12File(SGD_P12_FILENAME, "00");
        PrivateKey sgdSignerKey = cryptoUtilities.loadPrivateKeyFromP12File(SGD_P12_FILENAME, "00", "00");
        PublicKey sgdPublicKey = cryptoUtilities.loadX509CertificateFromP12File(SGD_P12_FILENAME, "00").getPublicKey();

        GetPublicKeyResponse sgdResponse = GetPublicKeyResponse.builder()
                .certificate(cryptoUtilities.convertBytesToB64String(certificate.getEncoded()))
                .signature(cryptoUtilities.convertBytesToB64String(cryptoUtilities.createSignatureUsingEcdsa(PUBLIC_KEY_SPEC.getBytes(), sgdSignerKey)))
                .publicKeyEcies(new PublicKeyEcies(PUBLIC_KEY_SPEC))
                .build();

        SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(sgdResponse, SIGNER_LIST, SgdInstance.SGD_1);
    }

    @Test
    public void signatureForDifferentContent_shouldThrowInvalidSignatureException() throws Exception {
        final X509Certificate certificate = cryptoUtilities.loadX509CertificateFromP12File(SGD_P12_FILENAME, "00");
        PrivateKey sgdSignerKey = cryptoUtilities.loadPrivateKeyFromP12File(SGD_P12_FILENAME, "00", "00");

        GetPublicKeyResponse sgdResponse = GetPublicKeyResponse.builder()
                .certificate(cryptoUtilities.convertBytesToB64String(certificate.getEncoded()))
                .signature(cryptoUtilities.convertBytesToB64String(cryptoUtilities.createSignatureUsingEcdsa((PUBLIC_KEY_SPEC + "0").getBytes(), sgdSignerKey)))
                .publicKeyEcies(new PublicKeyEcies(PUBLIC_KEY_SPEC))
                .build();

        assertThatThrownBy(() -> SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(sgdResponse, SIGNER_LIST, SgdInstance.SGD_1))
                .isInstanceOf(InvalidSignatureException.class);
    }
}
