/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.AesGcmEncryptionData;
import de.gematik.ti.sgd.domain.SgdInstance;
import de.gematik.ti.sgd.domain.json.GetPublicKeyResponse;
import de.gematik.ti.sgd.domain.json.SgdCommand;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.domain.sgdCommunication.EncryptedMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import de.gematik.ti.sgd.domain.sgdCommunication.vectors.SgdVector;
import de.gematik.ti.sgd.exceptions.server.ServerInvalidCertificateException;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import org.apache.commons.codec.digest.DigestUtils;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateHolder;
import org.bouncycastle.cert.ocsp.*;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.DigestCalculatorProvider;
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import org.junit.Before;
import org.junit.Test;
import org.springframework.test.util.ReflectionTestUtils;

import javax.xml.bind.DatatypeConverter;
import java.io.File;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

public class SchluesselGenerierungsDienstTest {
    private static List<PublicKey> SERVER_SIGNER_LIST;
    private static List<PublicKey> CLIENT_SIGNER_LIST;
    private static List<X509Certificate> OCSP_SIGNER_LIST;

    private static SgdCryptoUtilities cryptoUtilities = new SgdTestCryptoUtilities();

    static {
        SgdCryptoUtilities.setInstance(new SgdTestCryptoUtilities());

        SERVER_SIGNER_LIST = extractPublicKeysFromListOfCertificateFilenames(Arrays.asList("src/test/resources/keys/ca/GEM.EGK-CA10-TEST-ONLY.crt"));

        CLIENT_SIGNER_LIST = extractPublicKeysFromListOfCertificateFilenames(Arrays.asList("src/test/resources/keys/ca/GEM.KOMP-CA10-TEST-ONLY.crt"));

        OCSP_SIGNER_LIST = Stream.of("src/test/resources/keys/ocsp/TEST20190723_c.gem.ocsp.crt",
                "src/test/resources/keys/ocsp/TEST20190723_c.gem.ocsp-ecc.crt",
                "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.crt")
                .map(filename -> {
                    try {
                        return Optional.of(cryptoUtilities.loadX509CertificateFromCrtFile(filename));
                    } catch (Exception e) {
                        return Optional.empty();
                    }
                })
                .filter(Optional::isPresent)
                .map(Optional::get)
                .map(o -> (X509Certificate) o)
                .collect(Collectors.toList());
    }

    private final String ECC_AUT_CERTIFICATE_FILE = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.crt";
    private final String ECC_AUT_KEY_FILE = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.p12";
    private final String SGD1_P12_FILENAME = "src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12";
    private final String SGD2_P12_FILENAME = "src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12";
    private final String OCSP_SIGNER_P12_FILENAME = "src/test/resources/keys/ocsp/TEST20190723_c.gem.ocsp-ecc.p12";
    private final String OCSP_CA_CRT_FILENAME = "src/test/resources/keys/ca/GEM.OCSP-CA9-TEST-ONLY.crt";

    // system under test
    private SchluesselGenerierungsDienst sgd1;
    private SchluesselGenerierungsDienst sgd2;
    private X509Certificate sgd1Certificate;
    private KeyPair sgd1KeyPair;
    private X509Certificate sgd2Certificate;
    private KeyPair sgd2KeyPair;
    private X509Certificate clientCertificate;
    private KeyPair clientKeyPair;
    private X509Certificate ocspRsaSignerCertificate;
    private KeyPair ocspRsaSignerKey;
    private X509Certificate ocspRsaCaCertificate;

    public static OCSPResp gen(X509Certificate eeCertJca,
                               X509Certificate caCertJca,
                               PrivateKey caPrivateKey,
                               CertificateStatus status,
                               ZonedDateTime dateTime) throws CertificateException {
        try {
            X509CertificateHolder caCert = new JcaX509CertificateHolder(caCertJca);

            DigestCalculatorProvider digCalcProv = new BcDigestCalculatorProvider();
            BasicOCSPRespBuilder basicBuilder = new BasicOCSPRespBuilder(
                    SubjectPublicKeyInfo.getInstance(caCertJca.getPublicKey().getEncoded()),
                    digCalcProv.get(CertificateID.HASH_SHA1));

            CertificateID certId = new CertificateID(digCalcProv.get(CertificateID.HASH_SHA1),
                    caCert, eeCertJca.getSerialNumber());

            basicBuilder.addResponse(certId, status);

            X509CertificateHolder[] chain = {new X509CertificateHolder(caCertJca.getEncoded())};
            BasicOCSPResp resp = basicBuilder.build(
                    new JcaContentSignerBuilder("SHA256withECDSA").setProvider(BouncyCastleProvider.PROVIDER_NAME).build(caPrivateKey),
                    chain, new Date(dateTime.toInstant().toEpochMilli()));

            OCSPRespBuilder builder = new OCSPRespBuilder();
            return builder.build(OCSPRespBuilder.SUCCESSFUL, resp);
        } catch (Exception e) {
            throw new CertificateException("cannot generate OCSP response", e);
        }
    }

    @Before
    public void initSgd() {
        sgd1 = new SchluesselGenerierungsDienst();
        sgd2 = new SchluesselGenerierungsDienst();

        sgd1Certificate = cryptoUtilities.loadX509CertificateFromP12File(SGD1_P12_FILENAME, "00");
        sgd1KeyPair = cryptoUtilities.loadKeypairFromP12File(SGD1_P12_FILENAME, "00", "00");
        sgd2Certificate = cryptoUtilities.loadX509CertificateFromP12File(SGD2_P12_FILENAME, "00");
        sgd2KeyPair = cryptoUtilities.loadKeypairFromP12File(SGD2_P12_FILENAME, "00", "00");
        clientCertificate = cryptoUtilities.loadX509CertificateFromCrtFile(ECC_AUT_CERTIFICATE_FILE);
        clientKeyPair = cryptoUtilities.loadKeypairFromP12File(ECC_AUT_KEY_FILE, "00", "00");

        ocspRsaSignerCertificate = cryptoUtilities.loadX509CertificateFromP12File(OCSP_SIGNER_P12_FILENAME, "00");
        ocspRsaSignerKey = cryptoUtilities.loadKeypairFromP12File(OCSP_SIGNER_P12_FILENAME, "00", "00");
        ocspRsaCaCertificate = cryptoUtilities.loadX509CertificateFromCrtFile(OCSP_CA_CRT_FILENAME);

        ReflectionTestUtils.setField(sgd1, "sgdCertificate", sgd1Certificate);
        ReflectionTestUtils.setField(sgd1, "sgdPrivateKey", cryptoUtilities.loadPrivateKeyFromP12File(SGD1_P12_FILENAME, "00", "00"));
        ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", SERVER_SIGNER_LIST);
        ReflectionTestUtils.setField(sgd1, "possibleOcspSignerCertificates", OCSP_SIGNER_LIST);
        ReflectionTestUtils.setField(sgd1, "isThisSgdOne", true);

        ReflectionTestUtils.setField(sgd2, "sgdCertificate", sgd2Certificate);
        ReflectionTestUtils.setField(sgd2, "sgdPrivateKey", cryptoUtilities.loadPrivateKeyFromP12File(SGD2_P12_FILENAME, "00", "00"));
        ReflectionTestUtils.setField(sgd2, "possibleCertificateSigners", SERVER_SIGNER_LIST);
        ReflectionTestUtils.setField(sgd2, "possibleOcspSignerCertificates", OCSP_SIGNER_LIST);
        ReflectionTestUtils.setField(sgd2, "isThisSgdOne", false);
    }

    @Test
    public void getPublicKey_HappyPath() {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);

        final GetPublicKeyResponse getPublicKeyResponse = sgd1.getPublicKey(request);

        assertThat(getPublicKeyResponse.getCertificate()).isNotBlank();
        assertThat(cryptoUtilities.convertB64StringToBytes(getPublicKeyResponse.getCertificate()))
                .isNotEmpty();
        assertThat(cryptoUtilities.convertB64StringToBytes(getPublicKeyResponse.getSignature()))
                .isNotEmpty();
        assertThat(getPublicKeyResponse.getPublicKeyEcies().getMessage())
                .isNotEmpty()
                .contains(" 0x");

        SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(getPublicKeyResponse, CLIENT_SIGNER_LIST, SgdInstance.SGD_1);
    }

    @Test
    public void defunctCertificate_ShouldThrowException() {
        SgdRequest request = SgdRequest.builder()
                .command(SgdCommand.GET_PUBLIC_KEY)
                .certificate("1234")
                .build();

        assertThatThrownBy(() -> sgd1.getPublicKey(request))
                .isInstanceOf(ServerInvalidCertificateException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void nonTrustedCertificate_ShouldThrowException() throws CertificateEncodingException {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
        request.setCertificate(cryptoUtilities.convertBytesToB64String(sgd1Certificate.getEncoded()));

        assertThatThrownBy(() -> sgd1.getPublicKey(request))
                .isInstanceOf(ServerInvalidCertificateException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void eciesFail_expectDecryptionFailException() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));
        SgdRequest request = clientMachine.getAuthenticationTokenRequest_sgd1();

        final String[] contentParts = request.getEncryptedMessage().getContent().split(" ");
        contentParts[5] = contentParts[5].toLowerCase();
        request.setEncryptedMessage(new EncryptedMessage(String.join(" ", contentParts)));

        assertThatThrownBy(() -> sgd1.getAuthenticationToken(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.DECRYPTION_FAIL);
    }

    @Test
    public void signatureInvalid_expectSignatureNotValidException() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));
        SgdRequest request = clientMachine.getAuthenticationTokenRequest_sgd1();
        request.setSignature(request.getSignature().toLowerCase());

        assertThatThrownBy(() -> sgd1.getAuthenticationToken(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.SIGNATURE_NOT_VALID);
    }

    @Test
    public void certificateNotAutCertInGetAuthenticationToken_expectCertificateNotValidException() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));

        ReflectionTestUtils.setField(clientMachine, "clientCertificate", sgd1Certificate);
        ReflectionTestUtils.setField(clientMachine, "clientCertificateKeyPair", sgd1KeyPair);
        ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", CLIENT_SIGNER_LIST);

        SgdRequest request = clientMachine.getAuthenticationTokenRequest_sgd1();

        assertThatThrownBy(() -> sgd1.getAuthenticationToken(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void certificateNotAutCertInKeyDerivation_expectCertificateNotValidException() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receiveAuthenticationTokenResponse_sgd1(sgd1.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd1()));
        clientMachine.receiveAuthenticationTokenResponse_sgd2(sgd2.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd2()));

        ReflectionTestUtils.setField(clientMachine, "clientCertificate", sgd1Certificate);
        ReflectionTestUtils.setField(clientMachine, "clientCertificateKeyPair", sgd1KeyPair);
        ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", CLIENT_SIGNER_LIST);

        SgdRequest request = clientMachine.getKeyDerivationRequest_Sgd1();
        assertThatThrownBy(() -> sgd1.keyDerivation(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void signatureInvalidInKeyDerivation_expectSignatureNotValidException() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receiveAuthenticationTokenResponse_sgd1(sgd1.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd1()));
        clientMachine.receiveAuthenticationTokenResponse_sgd2(sgd2.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd2()));

        SgdRequest request = clientMachine.getKeyDerivationRequest_Sgd1();
        request.setSignature(request.getSignature().toLowerCase());

        assertThatThrownBy(() -> sgd1.keyDerivation(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.SIGNATURE_NOT_VALID);
    }

    @Test
    public void triggerEciesKeyChange_keyShouldBeDifferent() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        final SgdRequest publicKeyRequest = clientMachine.getPublicKeyRequest();

        final GetPublicKeyResponse publicKeyResponse1 = sgd1.getPublicKey(publicKeyRequest);
        final GetPublicKeyResponse publicKeyResponse2 = sgd1.getPublicKey(publicKeyRequest);

        sgd1.performEciesKeyChange();

        final GetPublicKeyResponse publicKeyResponse3 = sgd1.getPublicKey(publicKeyRequest);

        assertThat(publicKeyResponse1.getPublicKeyEcies()).isEqualTo(publicKeyResponse2.getPublicKeyEcies());
        assertThat(publicKeyResponse1.getPublicKeyEcies()).isNotEqualTo(publicKeyResponse3.getPublicKeyEcies());
    }

    @Test
    public void triggerEciesKeyChange_publicKeyFromBeforeShouldStillBeValid() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        assertThat(clientMachine.getSgd1AesGcmKey()).isNull();
        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));

        sgd1.performEciesKeyChange();

        clientMachine.receiveAuthenticationTokenResponse_sgd1(sgd1.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd1()));
        clientMachine.receiveAuthenticationTokenResponse_sgd2(sgd2.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd2()));
        clientMachine.receiveKeyDerivationResponse_sgd1(sgd1.keyDerivation(clientMachine.getKeyDerivationRequest_Sgd1()));
        clientMachine.receiveKeyDerivationResponse_sgd2(sgd2.keyDerivation(clientMachine.getKeyDerivationRequest_Sgd2()));

        assertThat(clientMachine.getSgd1AesGcmKey()).isNotNull();
    }

    @Test
    public void triggerEciesKeyChange_authenticationTokenShouldBeDifferent() {
        byte[] token1 = (byte[]) ReflectionTestUtils.getField(sgd1, "currentAuthenticationTokenDerivationKey");

        sgd1.performEciesKeyChange();

        byte[] token2 = (byte[]) ReflectionTestUtils.getField(sgd1, "currentAuthenticationTokenDerivationKey");

        assertThat(token1).isNotEqualTo(token2);
    }

    @Test
    public void wrongEcCurveInClient_ServerShouldGiveError() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);
        ReflectionTestUtils.setField(clientMachine, "ephemeralEciesKeyPair", SgdCryptoUtilities.getInstance().generateEccKeyPair("brainpoolp256t1"));

        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));

        assertThatThrownBy(() -> sgd1.getAuthenticationToken(clientMachine.getAuthenticationTokenRequest_sgd1()))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.PUBLIC_KEY_CORRUPTED);
    }

    @Test
    public void wrongHashInPublicKeyFieldInRequest_serverShouldGiveError() {
        SgdClientKeyDerivationMachine clientMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, CLIENT_SIGNER_LIST);

        clientMachine.receivePublicKeyResponse_sgd1(sgd1.getPublicKey(clientMachine.getPublicKeyRequest()));
        clientMachine.receivePublicKeyResponse_sgd2(sgd2.getPublicKey(clientMachine.getPublicKeyRequest()));

        final SgdRequest authenticationTokenRequest = clientMachine.getAuthenticationTokenRequest_sgd1();
        String[] parts = authenticationTokenRequest.getPublicKeyEcies().getMessage().split(" ");
        parts[3] = DigestUtils.sha256Hex(DatatypeConverter.parseHexBinary(parts[3]));
        parts[4] = DigestUtils.sha256Hex(DatatypeConverter.parseHexBinary(parts[4]));
        authenticationTokenRequest.setPublicKeyEcies(new PublicKeyEcies(String.join(" ", parts)));

        assertThatThrownBy(() -> sgd1.getAuthenticationToken(authenticationTokenRequest))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.SIGNATURE_NOT_VALID);
    }

    @Test
    public void getPublicKeyWithCorrectOcsp_ShouldReturnPublicKey() {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
        request.setOcspResponse(generateOcspResponse(clientCertificate, ocspRsaSignerCertificate, ocspRsaSignerKey.getPrivate(), ZonedDateTime.now()));

        final GetPublicKeyResponse getPublicKeyResponse = sgd1.getPublicKey(request);

        SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(getPublicKeyResponse, CLIENT_SIGNER_LIST, SgdInstance.SGD_1);
    }

    @Test
    public void getPublicKeyWithOcspWithWrongSigner_ShouldGiveException() {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
        request.setOcspResponse(generateOcspResponse(clientCertificate, sgd1Certificate, sgd1KeyPair.getPrivate(), ZonedDateTime.now()));

        assertThatThrownBy(() -> sgd1.getPublicKey(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void getPublicKeyWithOcspThatIsTooOld_ShouldGiveException() {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
        request.setOcspResponse(generateOcspResponse(clientCertificate, ocspRsaSignerCertificate, ocspRsaSignerKey.getPrivate(), ZonedDateTime.now().minusHours(7l)));

        assertThatThrownBy(() -> sgd1.getPublicKey(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void getPublicKeyWithOcspThatIsSignedByCaCert_ShouldGiveException() {
        SgdRequest request = SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
        request.setOcspResponse(generateOcspResponse(clientCertificate, clientCertificate, clientKeyPair.getPrivate(), ZonedDateTime.now()));

        assertThatThrownBy(() -> sgd1.getPublicKey(request))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void repeatedSgdKeyGeneration_keysShouldMatch() {
        final AesGcmEncryptionData encryptionData1 = sgd1.offlineKeyDerivationForKvnr("kvnr_value");
        final AesGcmEncryptionData encryptionData2 = sgd1.offlineKeyDerivationForKvnr(new SgdVector(encryptionData1.getAssociatedData()));

        assertThat(encryptionData1.getAesKey())
                .containsExactly(encryptionData2.getAesKey());
    }

    private String generateOcspResponse(X509Certificate clientCertificate, X509Certificate signerCertificate, PrivateKey signerKey, ZonedDateTime date) {
        try {
            OCSPResp ocspResp = gen(
                    clientCertificate,
                    signerCertificate,
                    signerKey,
                    CertificateStatus.GOOD,
                    date
            );

            return DatatypeConverter.printBase64Binary(ocspResp.getEncoded());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
